/*
	Asm versions of Xen pv-ops, suitable for either direct use or inlining.
	The inline versions are the same as the direct-use versions, with the
	pre- and post-amble chopped off.

	This code is encoded for size rather than absolute efficiency,
	with a view to being able to inline as much as possible.

	We only bother with direct forms (ie, vcpu in pda) of the operations
	here; the indirect forms are better handled in C, since they're
	generally too large to inline anyway.
 */

#include <linux/linkage.h>

#include <asm/asm-offsets.h>
#include <asm/thread_info.h>
#include <asm/percpu.h>
#include <asm/processor-flags.h>
#include <asm/segment.h>
#include <asm/asm-hack.h>

#include <xen/interface/xen.h>
#include <xen/interface/features.h>

#include <asm/calling.h>
#include <asm/processor-flags.h>


#define RELOC(x, v)	.globl x##_reloc; x##_reloc=v
#define ENDPATCH(x)	.globl x##_end; x##_end=.


/* Pseudo-flag used for virtual NMI, which we don't implement yet */
#define XEN_EFLAGS_NMI	0x80000000

/*
	Enable events.  This clears the event mask and tests the pending
	event status with one and operation.  If there are pending
	events, then enter the hypervisor to get them handled.
 */
ENTRY(xen_irq_enable_direct)
	/* Unmask events */
#ifdef PER_CPU_VAR
	movb $0, PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_mask
#else
	PER_CPU(xen_vcpu_info, rAX)
	movb $0, XEN_vcpu_info_mask(rAX)
#endif

	/* Preempt here doesn't matter because that will deal with
	   any pending interrupts.  The pending check may end up being
	   run on the wrong CPU, but that doesn't hurt. */

	/* Test for pending */
#ifdef PER_CPU_VAR
	testb $0xff, PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_pending
#else
	/* rAX already points to xen_vcpu_info */
	testb $0xff, XEN_vcpu_info_pending(rAX)
#endif
	jz 1f

2:	call check_events
1:
ENDPATCH(xen_irq_enable_direct)
	ret
	ENDPROC(xen_irq_enable_direct)
	RELOC(xen_irq_enable_direct, 2b+1)


/*
	Disabling events is simply a matter of making the event mask
	non-zero.
 */
ENTRY(xen_irq_disable_direct)
#ifdef PER_CPU_VAR
	movb $1, PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_mask
#else
	PER_CPU(xen_vcpu_info, rAX)
	movb $1, XEN_vcpu_info_mask(rAX)
#endif
ENDPATCH(xen_irq_disable_direct)
	ret
	ENDPROC(xen_irq_disable_direct)
	RELOC(xen_irq_disable_direct, 0)

/*
	(xen_)save_fl is used to get the current interrupt enable status.
	Callers expect the status to be in X86_EFLAGS_IF, and other bits
	may be set in the return value.  We take advantage of this by
	making sure that X86_EFLAGS_IF has the right value (and other bits
	in that byte are 0), but other bits in the return value are
	undefined.  We need to toggle the state of the bit, because
	Xen and x86 use opposite senses (mask vs enable).
 */
ENTRY(xen_save_fl_direct)
#ifdef PER_CPU_VAR
	testb $0xff, PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_mask
#else
	PER_CPU(xen_vcpu_info, rAX)
	testb $0xff, XEN_vcpu_info_mask(rAX)
#endif
	setz %ah
	addb %ah,%ah
ENDPATCH(xen_save_fl_direct)
	ret
	ENDPROC(xen_save_fl_direct)
	RELOC(xen_save_fl_direct, 0)


/*
	In principle the caller should be passing us a value return
	from xen_save_fl_direct, but for robustness sake we test only
	the X86_EFLAGS_IF flag rather than the whole byte. After
	setting the interrupt mask state, it checks for unmasked
	pending events and enters the hypervisor to get them delivered
	if so.
 */
ENTRY(xen_restore_fl_direct)
	testb $X86_EFLAGS_IF>>8, %ah
#ifdef PER_CPU_VAR
	setz PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_mask
#else
	PER_CPU(xen_vcpu_info, rAX)
	setz XEN_vcpu_info_mask(rAX)
#endif
	/* Preempt here doesn't matter because that will deal with
	   any pending interrupts.  The pending check may end up being
	   run on the wrong CPU, but that doesn't hurt. */

	/* check for unmasked and pending */
#ifdef PER_CPU_VAR
	cmpw $0x0001, PER_CPU_VAR(xen_vcpu_info)+XEN_vcpu_info_pending
#else
	/* rAX already points to xen_vcpu_info */
	cmpw $0x0001, XEN_vcpu_info_pending(rAX)
#endif
	jz 1f
2:	call check_events
1:
ENDPATCH(xen_restore_fl_direct)
	ret
	ENDPROC(xen_restore_fl_direct)
	RELOC(xen_restore_fl_direct, 2b+1)


#ifdef CONFIG_X86_64
#include "xen-asm_64.S"
#else
#include "xen-asm_32.S"
#endif
