#ifndef _X86_64_PGALLOC_H
#define _X86_64_PGALLOC_H

#include <asm/pda.h>
#include <asm/vsyscall.h>
#include <linux/threads.h>
#include <linux/mm.h>

#define arch_add_exec_range(mm, limit) \
		do { (void)(mm), (void)(limit); } while (0)
#define arch_flush_exec_range(mm) \
		do { (void)(mm); } while (0)
#define arch_remove_exec_range(mm, limit) \
		do { (void)(mm), (void)(limit); } while (0)

#ifdef CONFIG_XEN

#include <xen/page.h>

extern pud_t level3_user_pgt[512];

/*FIXME: use pvops */
static pgd_t *__user_pgd(pgd_t *pgd)
{
	return pgd + PTRS_PER_PGD;
}

#endif

#define pmd_populate_kernel(mm, pmd, pte) do { \
		paravirt_alloc_pt(mm, __pa(pte) >> PAGE_SHIFT); \
		set_pmd(pmd, __pmd(_PAGE_TABLE | __pa(pte))); \
		} while (0)
#define pud_populate(mm, pud, pmd) do { \
		paravirt_alloc_pd(mm, __pa(pmd) >> PAGE_SHIFT); \
		set_pud(pud, __pud(_PAGE_TABLE | __pa(pmd))); \
		} while (0)
#define pgd_populate(mm, pgd, pud) do {\
		paravirt_alloc_pd(mm, __pa(pud) >> PAGE_SHIFT); \
		set_pgd(pgd, __pgd(_PAGE_TABLE | __pa(pud))); \
		set_pgd(__user_pgd(pgd), __pgd(_PAGE_TABLE | __pa(pud))); \
		} while (0)

#define pmd_pgtable(pmd) pmd_page(pmd)

static inline void pmd_populate(struct mm_struct *mm, pmd_t *pmd, struct page *pte)
{
	paravirt_alloc_pt(mm, page_to_pfn(pte)); \
	set_pmd(pmd, __pmd(_PAGE_TABLE | (page_to_pfn(pte) << PAGE_SHIFT)));
}

static inline void pmd_free(struct mm_struct *mm, pmd_t *pmd)
{
	BUG_ON((unsigned long)pmd & (PAGE_SIZE-1));
	free_page((unsigned long)pmd);
}

static inline pmd_t *pmd_alloc_one (struct mm_struct *mm, unsigned long addr)
{
	pmd_t *pmd = (pmd_t *)get_zeroed_page(GFP_KERNEL|__GFP_REPEAT);
	paravirt_alloc_pd(mm, __pa(pmd) >> PAGE_SHIFT);
	return pmd;
}

static inline pud_t *pud_alloc_one(struct mm_struct *mm, unsigned long addr)
{
	pud_t *pud = (pud_t *)get_zeroed_page(GFP_KERNEL|__GFP_REPEAT);
	paravirt_alloc_pd(mm, __pa(pud) >> PAGE_SHIFT);
	return pud;
}

static inline void pud_free(struct mm_struct *mm, pud_t *pud)
{
	BUG_ON((unsigned long)pud & (PAGE_SIZE-1));
	free_page((unsigned long)pud);
}

static inline void pgd_list_add(pgd_t *pgd)
{
	struct page *page = virt_to_page(pgd);
	unsigned long flags;

	spin_lock_irqsave(&pgd_lock, flags);
	list_add(&page->lru, &pgd_list);
	spin_unlock_irqrestore(&pgd_lock, flags);
}

static inline void pgd_list_del(pgd_t *pgd)
{
	struct page *page = virt_to_page(pgd);
	unsigned long flags;

	spin_lock_irqsave(&pgd_lock, flags);
	list_del(&page->lru);
	spin_unlock_irqrestore(&pgd_lock, flags);
}

static inline pgd_t *pgd_alloc(struct mm_struct *mm)
{
	unsigned boundary;
	pgd_t *pgd = (pgd_t *)__get_free_pages(GFP_KERNEL|__GFP_REPEAT, 1);
	if (!pgd)
		return NULL;
	pgd_list_add(pgd);
	/*
	 * Copy kernel pointers in from init.
	 * Could keep a freelist or slab cache of those because the kernel
	 * part never changes.
	 */
	boundary = pgd_index(__PAGE_OFFSET);
	memset(pgd, 0, boundary * sizeof(pgd_t));
	memcpy(pgd + boundary,
	       init_level4_pgt + boundary,
	       (PTRS_PER_PGD - boundary) * sizeof(pgd_t));

	memset(__user_pgd(pgd), 0, PAGE_SIZE); /* clean up user pgd */
	/*
	 * Set level3_user_pgt for vsyscall area
	 */
	set_pgd(__user_pgd(pgd) + pgd_index(VSYSCALL_START), 
		mk_kernel_pgd(__pa_symbol(level3_user_pgt)));

	return pgd;
}

static inline void pgd_free(struct mm_struct *mm, pgd_t *pgd)
{
	BUG_ON((unsigned long)pgd & (PAGE_SIZE-1));
	pgd_list_del(pgd);

#ifdef CONFIG_XEN
	/* Unpin and make it read-write again, in the (likely) case
	 * it was pinned
	 *
	 * FIXME: should be a paravirt_ops hook
	 */
	xen_pgd_unpin(pgd);
	make_lowmem_page_readwrite(pgd);
#endif

	free_pages((unsigned long)pgd, 1);
}

static inline pte_t *pte_alloc_one_kernel(struct mm_struct *mm, unsigned long address)
{
	return (pte_t *)get_zeroed_page(GFP_KERNEL|__GFP_REPEAT);
}

static inline pgtable_t pte_alloc_one(struct mm_struct *mm, unsigned long address)
{
	struct page *page;
	void *p;

	p = (void *)get_zeroed_page(GFP_KERNEL|__GFP_REPEAT);
	if (!p)
		return NULL;
	page = virt_to_page(p);
	pgtable_page_ctor(page);
	return page;
}

/* Should really implement gc for free page table pages. This could be
   done with a reference count in struct page. */

static inline void pte_free_kernel(struct mm_struct *mm, pte_t *pte)
{
	BUG_ON((unsigned long)pte & (PAGE_SIZE-1));
	free_page((unsigned long)pte); 
}

static inline void pte_free(struct mm_struct *mm, pgtable_t pte)
{
	pgtable_page_dtor(pte);
	__free_page(pte);
} 

#define __pte_free_tlb(tlb,pte)				\
do {							\
	pgtable_page_dtor((pte));				\
	paravirt_release_pt(page_to_pfn(pte));		\
	tlb_remove_page((tlb), (pte));			\
} while (0)

#define __pmd_free_tlb(tlb,x)				\
do {							\
	paravirt_release_pd(page_to_pfn(virt_to_page(x)));	\
	tlb_remove_page((tlb),virt_to_page(x));			\
} while (0)

#define __pud_free_tlb(tlb,x)				\
do {							\
	paravirt_release_pd(page_to_pfn(virt_to_page(x)));	\
	tlb_remove_page((tlb),virt_to_page(x));			\
} while (0)

#endif /* _X86_64_PGALLOC_H */
